@@ -1,5 +1,5 @@
 # mssql/base.py
-# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -9,6 +9,16 @@
     :name: Microsoft SQL Server
 
 
+.. _mssql_external_dialects:
+
+External Dialects
+-----------------
+
+In addition to the above DBAPI layers with native SQLAlchemy support, there
+are third-party dialects for other DBAPI layers that are compatible
+with SQL Server. See the "External Dialects" list on the
+:ref:`dialect_toplevel` page. 
+
 .. _mssql_identity:
 
 Auto Increment Behavior / IDENTITY Columns
@@ -2043,14 +2053,6 @@ class MSDDLCompiler(compiler.DDLCompiler):
             ),
         )
 
-        whereclause = index.dialect_options["mssql"]["where"]
-
-        if whereclause is not None:
-            where_compiled = self.sql_compiler.process(
-                whereclause, include_table=False, literal_binds=True
-            )
-            text += " WHERE " + where_compiled
-
         # handle other included columns
         if index.dialect_options["mssql"]["include"]:
             inclusions = [
@@ -2064,6 +2066,14 @@ class MSDDLCompiler(compiler.DDLCompiler):
                 [preparer.quote(c.name) for c in inclusions]
             )
 
+        whereclause = index.dialect_options["mssql"]["where"]
+
+        if whereclause is not None:
+            where_compiled = self.sql_compiler.process(
+                whereclause, include_table=False, literal_binds=True
+            )
+            text += " WHERE " + where_compiled
+
         return text
 
     def visit_drop_index(self, drop):
@@ -2796,7 +2806,7 @@ class MSDialect(default.DefaultDialect):
                 C.c.table_name == tablename,
                 C.c.table_schema == owner,
             ),
-        )
+        ).order_by(TC.c.constraint_name, C.c.ordinal_position)
         c = connection.execute(s)
         constraint_name = None
         for row in c:
