@@ -1,12 +1,12 @@
 # engine/result.py
-# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
 # the MIT License: http://www.opensource.org/licenses/mit-license.php
 
-"""Define result set constructs including :class:`.ResultProxy`
-and :class:`.RowProxy."""
+"""Define result set constructs including :class:`_engine.ResultProxy`
+and :class:`.RowProxy`."""
 
 
 import collections
@@ -18,6 +18,7 @@ from ..sql import expression
 from ..sql import sqltypes
 from ..sql import util as sql_util
 
+
 # This reconstructor is necessary so that pickles with the C extension or
 # without use the same Binary format.
 try:
@@ -82,8 +83,8 @@ except ImportError:
         def __getitem__(self, key):
             try:
                 processor, obj, index = self._keymap[key]
-            except KeyError:
-                processor, obj, index = self._parent._key_fallback(key)
+            except KeyError as err:
+                processor, obj, index = self._parent._key_fallback(key, err)
             except TypeError:
                 if isinstance(key, slice):
                     l = []
@@ -111,17 +112,24 @@ except ImportError:
             try:
                 return self[name]
             except KeyError as e:
-                raise AttributeError(e.args[0])
+                util.raise_(AttributeError(e.args[0]), replace_context=e)
 
 
 class RowProxy(BaseRowProxy):
-    """Proxy values from a single cursor row.
+    """Represent a single result row.
+
+    The :class:`.RowProxy` object is retrieved from a database result, from the
+    :class:`_engine.ResultProxy` object using methods like
+    :meth:`_engine.ResultProxy.fetchall`.
+
+    The :class:`.RowProxy` object seeks to act mostly like a Python named
+    tuple, but also provides some Python dictionary behaviors at the same time.
+
+    .. seealso::
+
+        :ref:`coretutorial_selecting` - includes examples of selecting
+        rows from SELECT statements.
 
-    Mostly follows "ordered dictionary" behavior, mapping result
-    values to the string-based column name, the integer position of
-    the result in the row, as well as Column instances which can be
-    mapped to the original Columns that produced this result set (for
-    results that correspond to constructed SQL expressions).
     """
 
     __slots__ = ()
@@ -169,26 +177,73 @@ class RowProxy(BaseRowProxy):
         return repr(sql_util._repr_row(self))
 
     def has_key(self, key):
-        """Return True if this RowProxy contains the given key."""
+        """Return True if this :class:`.RowProxy` contains the given key.
+
+        Through the SQLAlchemy 1.x series, the ``__contains__()`` method
+        of :class:`.RowProxy` also links to :meth:`.RowProxy.has_key`, in that
+        an expression such as ::
+
+            "some_col" in row
+
+        Will return True if the row contains a column named ``"some_col"``,
+        in the way that a Python mapping works.
+
+        However, it is planned that the 2.0 series of SQLAlchemy will reverse
+        this behavior so that ``__contains__()`` will refer to a value being
+        present in the row, in the way that a Python tuple works.
+
+        """
 
         return self._parent._has_key(key)
 
     def items(self):
-        """Return a list of tuples, each tuple containing a key/value pair."""
-        # TODO: no coverage here
+        """Return a list of tuples, each tuple containing a key/value pair.
+
+        This method is analogous to the Python dictionary ``.items()`` method,
+        except that it returns a list, not an iterator.
+
+        """
+
         return [(key, self[key]) for key in self.keys()]
 
     def keys(self):
-        """Return the list of keys as strings represented by this RowProxy."""
+        """Return the list of keys as strings represented by this
+        :class:`.RowProxy`.
+
+        This method is analogous to the Python dictionary ``.keys()`` method,
+        except that it returns a list, not an iterator.
+
+        """
 
         return self._parent.keys
 
     def iterkeys(self):
+        """Return a an iterator against the :meth:`.RowProxy.keys` method.
+
+        This method is analogous to the Python-2-only dictionary
+        ``.iterkeys()`` method.
+
+        """
         return iter(self._parent.keys)
 
     def itervalues(self):
+        """Return a an iterator against the :meth:`.RowProxy.values` method.
+
+        This method is analogous to the Python-2-only dictionary
+        ``.itervalues()`` method.
+
+        """
         return iter(self)
 
+    def values(self):
+        """Return the values represented by this :class:`.RowProxy` as a list.
+
+        This method is analogous to the Python dictionary ``.values()`` method,
+        except that it returns a list, not an iterator.
+
+        """
+        return super(RowProxy, self).values()
+
 
 try:
     # Register RowProxy with Sequence,
@@ -219,9 +274,11 @@ class ResultMetaData(object):
         self._orig_processors = None
 
         if context.result_column_struct:
-            result_columns, cols_are_ordered, textual_ordered = (
-                context.result_column_struct
-            )
+            (
+                result_columns,
+                cols_are_ordered,
+                textual_ordered,
+            ) = context.result_column_struct
             num_ctx_cols = len(result_columns)
         else:
             result_columns = (
@@ -362,7 +419,7 @@ class ResultMetaData(object):
 
         The remaining fairly common case is that of the textual SQL
         that includes at least partial column information; this is when
-        we use a :class:`.TextAsFrom` construct.   This contruct may have
+        we use a :class:`.TextAsFrom` construct.   This construct may have
         unordered or ordered column information.  In the ordered case, we
         merge the cursor.description and the compiled construct's information
         positionally, and warn if there are additional description names
@@ -502,7 +559,6 @@ class ResultMetaData(object):
     def _merge_textual_cols_by_position(
         self, context, cursor_description, result_columns
     ):
-        dialect = context.dialect
         num_ctx_cols = len(result_columns) if result_columns else None
 
         if num_ctx_cols > len(cursor_description):
@@ -511,7 +567,6 @@ class ResultMetaData(object):
                 "smaller than number of columns requested (%d)"
                 % (num_ctx_cols, len(cursor_description))
             )
-
         seen = set()
         for (
             idx,
@@ -558,7 +613,6 @@ class ResultMetaData(object):
             yield idx, colname, mapped_type, coltype, obj, untranslated
 
     def _merge_cols_by_none(self, context, cursor_description):
-        dialect = context.dialect
         for (
             idx,
             colname,
@@ -585,7 +639,7 @@ class ResultMetaData(object):
                 d[key] = rec
         return d
 
-    def _key_fallback(self, key, raiseerr=True):
+    def _key_fallback(self, key, err, raiseerr=True):
         map_ = self._keymap
         result = None
         if isinstance(key, util.string_types):
@@ -624,9 +678,12 @@ class ResultMetaData(object):
                     result = None
         if result is None:
             if raiseerr:
-                raise exc.NoSuchColumnError(
-                    "Could not locate column in row for column '%s'"
-                    % expression._string_or_unprintable(key)
+                util.raise_(
+                    exc.NoSuchColumnError(
+                        "Could not locate column in row for column '%s'"
+                        % expression._string_or_unprintable(key)
+                    ),
+                    replace_context=err,
                 )
             else:
                 return None
@@ -638,21 +695,24 @@ class ResultMetaData(object):
         if key in self._keymap:
             return True
         else:
-            return self._key_fallback(key, False) is not None
+            return self._key_fallback(key, None, False) is not None
 
     def _getter(self, key, raiseerr=True):
         if key in self._keymap:
             processor, obj, index = self._keymap[key]
         else:
-            ret = self._key_fallback(key, raiseerr)
+            ret = self._key_fallback(key, None, raiseerr)
             if ret is None:
                 return None
             processor, obj, index = ret
 
         if index is None:
-            raise exc.InvalidRequestError(
-                "Ambiguous column name '%s' in "
-                "result set column descriptions" % obj
+            util.raise_(
+                exc.InvalidRequestError(
+                    "Ambiguous column name '%s' in "
+                    "result set column descriptions" % obj
+                ),
+                from_=None,
             )
 
         return operator.itemgetter(index)
@@ -684,23 +744,16 @@ class ResultMetaData(object):
 
 
 class ResultProxy(object):
-    """Wraps a DB-API cursor object to provide easier access to row columns.
+    """A facade around a DBAPI cursor object.
 
-    Individual columns may be accessed by their integer position,
-    case-insensitive column name, or by ``schema.Column``
-    object. e.g.::
+    Returns database rows via the :class:`.RowProxy` class, which provides
+    additional API features and behaviors on top of the raw data returned
+    by the DBAPI.
 
-      row = fetchone()
+    .. seealso::
 
-      col1 = row[0]    # access via integer position
-
-      col2 = row['col2']   # access via name
-
-      col3 = row[mytable.c.mycol] # access via Column object.
-
-    ``ResultProxy`` also handles post-processing of result column
-    data using ``TypeEngine`` objects, which are referenced from
-    the originating SQL statement that produced this result set.
+        :ref:`coretutorial_selecting` - introductory material for accessing
+        :class:`_engine.ResultProxy` and :class:`.RowProxy` objects.
 
     """
 
@@ -724,16 +777,16 @@ class ResultProxy(object):
     def _getter(self, key, raiseerr=True):
         try:
             getter = self._metadata._getter
-        except AttributeError:
-            return self._non_result(None)
+        except AttributeError as err:
+            return self._non_result(None, err)
         else:
             return getter(key, raiseerr)
 
     def _has_key(self, key):
         try:
             has_key = self._metadata._has_key
-        except AttributeError:
-            return self._non_result(None)
+        except AttributeError as err:
+            return self._non_result(None, err)
         else:
             return has_key(key)
 
@@ -758,7 +811,9 @@ class ResultProxy(object):
                 )
 
     def keys(self):
-        """Return the current set of string keys for rows."""
+        """Return the list of string keys that would represented by each
+        :class:`.RowProxy`."""
+
         if self._metadata:
             return self._metadata.keys
         else:
@@ -773,7 +828,7 @@ class ResultProxy(object):
 
         .. note::
 
-           Notes regarding :attr:`.ResultProxy.rowcount`:
+           Notes regarding :attr:`_engine.ResultProxy.rowcount`:
 
 
            * This attribute returns the number of rows *matched*,
@@ -786,18 +841,20 @@ class ResultProxy(object):
              rowcount is configured by default to return the match
              count in all cases.
 
-           * :attr:`.ResultProxy.rowcount` is *only* useful in conjunction
+           * :attr:`_engine.ResultProxy.rowcount`
+             is *only* useful in conjunction
              with an UPDATE or DELETE statement.  Contrary to what the Python
              DBAPI says, it does *not* return the
              number of rows available from the results of a SELECT statement
              as DBAPIs cannot support this functionality when rows are
              unbuffered.
 
-           * :attr:`.ResultProxy.rowcount` may not be fully implemented by
+           * :attr:`_engine.ResultProxy.rowcount`
+             may not be fully implemented by
              all dialects.  In particular, most DBAPIs do not support an
              aggregate rowcount result from an executemany call.
-             The :meth:`.ResultProxy.supports_sane_rowcount` and
-             :meth:`.ResultProxy.supports_sane_multi_rowcount` methods
+             The :meth:`_engine.ResultProxy.supports_sane_rowcount` and
+             :meth:`_engine.ResultProxy.supports_sane_multi_rowcount` methods
              will report from the dialect if each usage is known to be
              supported.
 
@@ -814,7 +871,7 @@ class ResultProxy(object):
 
     @property
     def lastrowid(self):
-        """return the 'lastrowid' accessor on the DBAPI cursor.
+        """Return the 'lastrowid' accessor on the DBAPI cursor.
 
         This is a DBAPI specific method and is only functional
         for those backends which support it, for statements
@@ -837,21 +894,21 @@ class ResultProxy(object):
 
     @property
     def returns_rows(self):
-        """True if this :class:`.ResultProxy` returns rows.
+        """True if this :class:`_engine.ResultProxy` returns rows.
 
         I.e. if it is legal to call the methods
-        :meth:`~.ResultProxy.fetchone`,
-        :meth:`~.ResultProxy.fetchmany`
-        :meth:`~.ResultProxy.fetchall`.
+        :meth:`_engine.ResultProxy.fetchone`,
+        :meth:`_engine.ResultProxy.fetchmany`
+        :meth:`_engine.ResultProxy.fetchall`.
 
         """
         return self._metadata is not None
 
     @property
     def is_insert(self):
-        """True if this :class:`.ResultProxy` is the result
+        """True if this :class:`_engine.ResultProxy` is the result
         of a executing an expression language compiled
-        :func:`.expression.insert` construct.
+        :func:`_expression.insert` construct.
 
         When True, this implies that the
         :attr:`inserted_primary_key` attribute is accessible,
@@ -867,7 +924,7 @@ class ResultProxy(object):
         return self._saved_cursor.description
 
     def _soft_close(self):
-        """Soft close this :class:`.ResultProxy`.
+        """Soft close this :class:`_engine.ResultProxy`.
 
         This releases all DBAPI cursor resources, but leaves the
         ResultProxy "open" from a semantic perspective, meaning the
@@ -885,7 +942,7 @@ class ResultProxy(object):
 
         .. seealso::
 
-            :meth:`.ResultProxy.close`
+            :meth:`_engine.ResultProxy.close`
 
 
         """
@@ -901,38 +958,45 @@ class ResultProxy(object):
     def close(self):
         """Close this ResultProxy.
 
-        This closes out the underlying DBAPI cursor corresonding
+        This closes out the underlying DBAPI cursor corresponding
         to the statement execution, if one is still present.  Note that the
-        DBAPI cursor is automatically released when the :class:`.ResultProxy`
-        exhausts all available rows.  :meth:`.ResultProxy.close` is generally
+        DBAPI cursor is automatically released when the
+        :class:`_engine.ResultProxy`
+        exhausts all available rows.  :meth:`_engine.ResultProxy.close`
+        is generally
         an optional method except in the case when discarding a
-        :class:`.ResultProxy` that still has additional rows pending for fetch.
+        :class:`_engine.ResultProxy`
+        that still has additional rows pending for fetch.
 
         In the case of a result that is the product of
         :ref:`connectionless execution <dbengine_implicit>`,
-        the underlying :class:`.Connection` object is also closed, which
+        the underlying :class:`_engine.Connection` object is also closed,
+        which
         :term:`releases` DBAPI connection resources.
 
         After this method is called, it is no longer valid to call upon
         the fetch methods, which will raise a :class:`.ResourceClosedError`
         on subsequent use.
 
-        .. versionchanged:: 1.0.0 - the :meth:`.ResultProxy.close` method
+        .. versionchanged:: 1.0.0 - the :meth:`_engine.ResultProxy.close`
+           method
            has been separated out from the process that releases the underlying
            DBAPI cursor resource.   The "auto close" feature of the
-           :class:`.Connection` now performs a so-called "soft close", which
+           :class:`_engine.Connection` now performs a so-called "soft close",
+           which
            releases the underlying DBAPI cursor, but allows the
-           :class:`.ResultProxy` to still behave as an open-but-exhausted
-           result set; the actual :meth:`.ResultProxy.close` method is never
-           called.    It is still safe to discard a :class:`.ResultProxy`
+           :class:`_engine.ResultProxy`
+           to still behave as an open-but-exhausted
+           result set; the actual :meth:`_engine.ResultProxy.close`
+           method is never
+           called.    It is still safe to discard a
+           :class:`_engine.ResultProxy`
            that has been fully exhausted without calling this method.
 
         .. seealso::
 
             :ref:`connections_toplevel`
 
-            :meth:`.ResultProxy._soft_close`
-
         """
 
         if not self.closed:
@@ -950,7 +1014,10 @@ class ResultProxy(object):
                 yield row
 
     def __next__(self):
-        """Implement the next() protocol.
+        """Implement the Python next() protocol.
+
+        This method, mirrored as both ``.next()`` and  ``.__next__()``, is part
+        of Python's API for producing iterator-like behavior.
 
         .. versionadded:: 1.2
 
@@ -971,14 +1038,14 @@ class ResultProxy(object):
         corresponding to the list of primary key columns
         in the target table.
 
-        This only applies to single row :func:`.insert`
+        This only applies to single row :func:`_expression.insert`
         constructs which did not explicitly specify
-        :meth:`.Insert.returning`.
+        :meth:`_expression.Insert.returning`.
 
         Note that primary key columns which specify a
         server_default clause,
         or otherwise do not qualify as "autoincrement"
-        columns (see the notes at :class:`.Column`), and were
+        columns (see the notes at :class:`_schema.Column`), and were
         generated using the database-side default, will
         appear in this list as ``None`` unless the backend
         supports "returning" and the insert statement executed
@@ -1128,7 +1195,7 @@ class ResultProxy(object):
     def supports_sane_rowcount(self):
         """Return ``supports_sane_rowcount`` from the dialect.
 
-        See :attr:`.ResultProxy.rowcount` for background.
+        See :attr:`_engine.ResultProxy.rowcount` for background.
 
         """
 
@@ -1137,7 +1204,7 @@ class ResultProxy(object):
     def supports_sane_multi_rowcount(self):
         """Return ``supports_sane_multi_rowcount`` from the dialect.
 
-        See :attr:`.ResultProxy.rowcount` for background.
+        See :attr:`_engine.ResultProxy.rowcount` for background.
 
         """
 
@@ -1146,8 +1213,8 @@ class ResultProxy(object):
     def _fetchone_impl(self):
         try:
             return self.cursor.fetchone()
-        except AttributeError:
-            return self._non_result(None)
+        except AttributeError as err:
+            return self._non_result(None, err)
 
     def _fetchmany_impl(self, size=None):
         try:
@@ -1155,23 +1222,29 @@ class ResultProxy(object):
                 return self.cursor.fetchmany()
             else:
                 return self.cursor.fetchmany(size)
-        except AttributeError:
-            return self._non_result([])
+        except AttributeError as err:
+            return self._non_result([], err)
 
     def _fetchall_impl(self):
         try:
             return self.cursor.fetchall()
-        except AttributeError:
-            return self._non_result([])
+        except AttributeError as err:
+            return self._non_result([], err)
 
-    def _non_result(self, default):
+    def _non_result(self, default, err=None):
         if self._metadata is None:
-            raise exc.ResourceClosedError(
-                "This result object does not return rows. "
-                "It has been closed automatically."
+            util.raise_(
+                exc.ResourceClosedError(
+                    "This result object does not return rows. "
+                    "It has been closed automatically."
+                ),
+                replace_context=err,
             )
         elif self.closed:
-            raise exc.ResourceClosedError("This result object is closed.")
+            util.raise_(
+                exc.ResourceClosedError("This result object is closed."),
+                replace_context=err,
+            )
         else:
             return default
 
@@ -1199,13 +1272,11 @@ class ResultProxy(object):
         cursor resource is released, and the object may be safely
         discarded.
 
-        Subsequent calls to :meth:`.ResultProxy.fetchall` will return
-        an empty list.   After the :meth:`.ResultProxy.close` method is
+        Subsequent calls to :meth:`_engine.ResultProxy.fetchall` will return
+        an empty list.   After the :meth:`_engine.ResultProxy.close` method is
         called, the method will raise :class:`.ResourceClosedError`.
 
-        .. versionchanged:: 1.0.0 - Added "soft close" behavior which
-           allows the result to be used in an "exhausted" state prior to
-           calling the :meth:`.ResultProxy.close` method.
+        :return: a list of :class:`.RowProxy` objects
 
         """
 
@@ -1226,14 +1297,13 @@ class ResultProxy(object):
         cursor resource is released, and the object may be safely
         discarded.
 
-        Calls to :meth:`.ResultProxy.fetchmany` after all rows have been
+        Calls to :meth:`_engine.ResultProxy.fetchmany`
+        after all rows have been
         exhausted will return
-        an empty list.   After the :meth:`.ResultProxy.close` method is
+        an empty list.   After the :meth:`_engine.ResultProxy.close` method is
         called, the method will raise :class:`.ResourceClosedError`.
 
-        .. versionchanged:: 1.0.0 - Added "soft close" behavior which
-           allows the result to be used in an "exhausted" state prior to
-           calling the :meth:`.ResultProxy.close` method.
+        :return: a list of :class:`.RowProxy` objects
 
         """
 
@@ -1254,14 +1324,12 @@ class ResultProxy(object):
         cursor resource is released, and the object may be safely
         discarded.
 
-        Calls to :meth:`.ResultProxy.fetchone` after all rows have
+        Calls to :meth:`_engine.ResultProxy.fetchone` after all rows have
         been exhausted will return ``None``.
-        After the :meth:`.ResultProxy.close` method is
+        After the :meth:`_engine.ResultProxy.close` method is
         called, the method will raise :class:`.ResourceClosedError`.
 
-        .. versionchanged:: 1.0.0 - Added "soft close" behavior which
-           allows the result to be used in an "exhausted" state prior to
-           calling the :meth:`.ResultProxy.close` method.
+        :return: a :class:`.RowProxy` object, or None if no rows remain
 
         """
         try:
@@ -1279,10 +1347,11 @@ class ResultProxy(object):
     def first(self):
         """Fetch the first row and then close the result set unconditionally.
 
-        Returns None if no row is present.
-
         After calling this method, the object is fully closed,
-        e.g. the :meth:`.ResultProxy.close` method will have been called.
+        e.g. the :meth:`_engine.ResultProxy.close`
+        method will have been called.
+
+        :return: a :class:`.RowProxy` object, or None if no rows remain
 
         """
         if self._metadata is None:
@@ -1306,10 +1375,11 @@ class ResultProxy(object):
     def scalar(self):
         """Fetch the first column of the first row, and close the result set.
 
-        Returns None if no row is present.
-
         After calling this method, the object is fully closed,
-        e.g. the :meth:`.ResultProxy.close` method will have been called.
+        e.g. the :meth:`_engine.ResultProxy.close`
+        method will have been called.
+
+        :return: a Python scalar value , or None if no rows remain
 
         """
         row = self.first()
@@ -1483,7 +1553,7 @@ class BufferedColumnResultProxy(ResultProxy):
     databases where result rows contain "live" results that fall out
     of scope unless explicitly fetched.
 
-    .. versionchanged:: 1.2  This :class:`.ResultProxy` is not used by
+    .. versionchanged:: 1.2  This :class:`_engine.ResultProxy` is not used by
        any SQLAlchemy-included dialects.
 
     """
