@@ -1,5 +1,5 @@
 # orm/descriptor_props.py
-# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -11,21 +11,28 @@ as actively in the load/persist ORM loop.
 
 """
 
-from .interfaces import MapperProperty, PropComparator
-from .util import _none_set
 from . import attributes
-from .. import util, sql, exc as sa_exc, event, schema
-from ..sql import expression
 from . import properties
 from . import query
+from .interfaces import MapperProperty
+from .interfaces import PropComparator
+from .util import _none_set
+from .. import event
+from .. import exc as sa_exc
+from .. import schema
+from .. import sql
+from .. import util
+from ..sql import expression
 
 
 class DescriptorProperty(MapperProperty):
     """:class:`.MapperProperty` which proxies access to a
-        user-defined descriptor."""
+    user-defined descriptor."""
 
     doc = None
 
+    uses_objects = False
+
     def instrument_class(self, mapper):
         prop = self
 
@@ -34,12 +41,18 @@ class DescriptorProperty(MapperProperty):
             expire_missing = True
             collection = False
 
+            @property
+            def uses_objects(self):
+                return prop.uses_objects
+
             def __init__(self, key):
                 self.key = key
 
-            if hasattr(prop, 'get_history'):
-                def get_history(self, state, dict_,
-                                passive=attributes.PASSIVE_OFF):
+            if hasattr(prop, "get_history"):
+
+                def get_history(
+                    self, state, dict_, passive=attributes.PASSIVE_OFF
+                ):
                     return prop.get_history(state, dict_, passive)
 
         if self.descriptor is None:
@@ -48,6 +61,7 @@ class DescriptorProperty(MapperProperty):
                 self.descriptor = desc
 
         if self.descriptor is None:
+
             def fset(obj, value):
                 setattr(obj, self.name, value)
 
@@ -57,26 +71,21 @@ class DescriptorProperty(MapperProperty):
             def fget(obj):
                 return getattr(obj, self.name)
 
-            self.descriptor = property(
-                fget=fget,
-                fset=fset,
-                fdel=fdel,
-            )
+            self.descriptor = property(fget=fget, fset=fset, fdel=fdel)
 
-        proxy_attr = attributes.create_proxied_attribute(
-            self.descriptor)(
-                self.parent.class_,
-                self.key,
-                self.descriptor,
-                lambda: self._comparator_factory(mapper),
-                doc=self.doc,
-                original_property=self
-            )
+        proxy_attr = attributes.create_proxied_attribute(self.descriptor)(
+            self.parent.class_,
+            self.key,
+            self.descriptor,
+            lambda: self._comparator_factory(mapper),
+            doc=self.doc,
+            original_property=self,
+        )
         proxy_attr.impl = _ProxyImpl(self.key)
         mapper.class_manager.instrument_attribute(self.key, proxy_attr)
 
 
-@util.langhelpers.dependency_for("sqlalchemy.orm.properties")
+@util.langhelpers.dependency_for("sqlalchemy.orm.properties", add_to_all=True)
 class CompositeProperty(DescriptorProperty):
     """Defines a "composite" mapped attribute, representing a collection
     of columns as one attribute.
@@ -90,6 +99,15 @@ class CompositeProperty(DescriptorProperty):
 
     """
 
+    @util.deprecated_params(
+        extension=(
+            "0.7",
+            ":class:`.AttributeExtension` is deprecated in favor of the "
+            ":class:`.AttributeEvents` listener interface.  The "
+            ":paramref:`.composite.extension` parameter will be "
+            "removed in a future release.",
+        )
+    )
     def __init__(self, class_, *attrs, **kwargs):
         r"""Return a composite column-based property for use with a Mapper.
 
@@ -100,7 +118,9 @@ class CompositeProperty(DescriptorProperty):
         is the :class:`.CompositeProperty`.
 
         :param class\_:
-          The "composite type" class.
+          The "composite type" class, or any classmethod or callable which
+          will produce a new instance of the composite object given the
+          column values in order.
 
         :param \*cols:
           List of Column objects to be mapped.
@@ -110,10 +130,6 @@ class CompositeProperty(DescriptorProperty):
           scalar attribute should be loaded when replaced, if not
           already loaded.  See the same flag on :func:`.column_property`.
 
-          .. versionchanged:: 0.7
-              This flag specifically becomes meaningful
-              - previously it was a placeholder.
-
         :param group:
           A group name for this property when marked as deferred.
 
@@ -134,26 +150,25 @@ class CompositeProperty(DescriptorProperty):
         :param info: Optional data dictionary which will be populated into the
             :attr:`.MapperProperty.info` attribute of this object.
 
-            .. versionadded:: 0.8
-
         :param extension:
           an :class:`.AttributeExtension` instance,
           or list of extensions, which will be prepended to the list of
           attribute listeners for the resulting descriptor placed on the
-          class.  **Deprecated.**  Please see :class:`.AttributeEvents`.
+          class.
 
         """
         super(CompositeProperty, self).__init__()
 
         self.attrs = attrs
         self.composite_class = class_
-        self.active_history = kwargs.get('active_history', False)
-        self.deferred = kwargs.get('deferred', False)
-        self.group = kwargs.get('group', None)
-        self.comparator_factory = kwargs.pop('comparator_factory',
-                                             self.__class__.Comparator)
-        if 'info' in kwargs:
-            self.info = kwargs.pop('info')
+        self.active_history = kwargs.get("active_history", False)
+        self.deferred = kwargs.get("deferred", False)
+        self.group = kwargs.get("group", None)
+        self.comparator_factory = kwargs.pop(
+            "comparator_factory", self.__class__.Comparator
+        )
+        if "info" in kwargs:
+            self.info = kwargs.pop("info")
 
         util.set_creation_order(self)
         self._create_descriptor()
@@ -184,8 +199,7 @@ class CompositeProperty(DescriptorProperty):
                 # attributes, retrieve their values.  This
                 # ensures they all load.
                 values = [
-                    getattr(instance, key)
-                    for key in self._attribute_keys
+                    getattr(instance, key) for key in self._attribute_keys
                 ]
 
                 # current expected behavior here is that the composite is
@@ -194,8 +208,7 @@ class CompositeProperty(DescriptorProperty):
                 # if the composite were created unconditionally,
                 # but that would be a behavioral change.
                 if self.key not in dict_ and (
-                    state.key is not None or
-                    not _none_set.issuperset(values)
+                    state.key is not None or not _none_set.issuperset(values)
                 ):
                     dict_[self.key] = self.composite_class(*values)
                     state.manager.dispatch.refresh(state, None, [self.key])
@@ -215,8 +228,8 @@ class CompositeProperty(DescriptorProperty):
                     setattr(instance, key, None)
             else:
                 for key, value in zip(
-                        self._attribute_keys,
-                        value.__composite_values__()):
+                    self._attribute_keys, value.__composite_values__()
+                ):
                     setattr(instance, key, value)
 
         def fdel(instance):
@@ -232,18 +245,14 @@ class CompositeProperty(DescriptorProperty):
 
     @util.memoized_property
     def _comparable_elements(self):
-        return [
-            getattr(self.parent.class_, prop.key)
-            for prop in self.props
-        ]
+        return [getattr(self.parent.class_, prop.key) for prop in self.props]
 
     @util.memoized_property
     def props(self):
         props = []
         for attr in self.attrs:
             if isinstance(attr, str):
-                prop = self.parent.get_property(
-                    attr, _configure_mappers=False)
+                prop = self.parent.get_property(attr, _configure_mappers=False)
             elif isinstance(attr, schema.Column):
                 prop = self.parent._columntoproperty[attr]
             elif isinstance(attr, attributes.InstrumentedAttribute):
@@ -252,7 +261,8 @@ class CompositeProperty(DescriptorProperty):
                 raise sa_exc.ArgumentError(
                     "Composite expects Column objects or mapped "
                     "attributes/attribute names as arguments, got: %r"
-                    % (attr,))
+                    % (attr,)
+                )
             props.append(prop)
         return props
 
@@ -269,9 +279,7 @@ class CompositeProperty(DescriptorProperty):
             prop.active_history = self.active_history
             if self.deferred:
                 prop.deferred = self.deferred
-                prop.strategy_key = (
-                    ("deferred", True),
-                    ("instrument", True))
+                prop.strategy_key = (("deferred", True), ("instrument", True))
             prop.group = self.group
 
     def _setup_event_handlers(self):
@@ -297,8 +305,7 @@ class CompositeProperty(DescriptorProperty):
                     return
 
             dict_[self.key] = self.composite_class(
-                *[state.dict[key] for key in
-                  self._attribute_keys]
+                *[state.dict[key] for key in self._attribute_keys]
             )
 
         def expire_handler(state, keys):
@@ -315,24 +322,27 @@ class CompositeProperty(DescriptorProperty):
 
             state.dict.pop(self.key, None)
 
-        event.listen(self.parent, 'after_insert',
-                     insert_update_handler, raw=True)
-        event.listen(self.parent, 'after_update',
-                     insert_update_handler, raw=True)
-        event.listen(self.parent, 'load',
-                     load_handler, raw=True, propagate=True)
-        event.listen(self.parent, 'refresh',
-                     refresh_handler, raw=True, propagate=True)
-        event.listen(self.parent, 'expire',
-                     expire_handler, raw=True, propagate=True)
+        event.listen(
+            self.parent, "after_insert", insert_update_handler, raw=True
+        )
+        event.listen(
+            self.parent, "after_update", insert_update_handler, raw=True
+        )
+        event.listen(
+            self.parent, "load", load_handler, raw=True, propagate=True
+        )
+        event.listen(
+            self.parent, "refresh", refresh_handler, raw=True, propagate=True
+        )
+        event.listen(
+            self.parent, "expire", expire_handler, raw=True, propagate=True
+        )
 
         # TODO: need a deserialize hook here
 
     @util.memoized_property
     def _attribute_keys(self):
-        return [
-            prop.key for prop in self.props
-        ]
+        return [prop.key for prop in self.props]
 
     def get_history(self, state, dict_, passive=attributes.PASSIVE_OFF):
         """Provided for userland code that uses attributes.get_history()."""
@@ -361,26 +371,27 @@ class CompositeProperty(DescriptorProperty):
             return attributes.History(
                 [self.composite_class(*added)],
                 (),
-                [self.composite_class(*deleted)]
+                [self.composite_class(*deleted)],
             )
         else:
-            return attributes.History(
-                (), [self.composite_class(*added)], ()
-            )
+            return attributes.History((), [self.composite_class(*added)], ())
 
     def _comparator_factory(self, mapper):
         return self.comparator_factory(self, mapper)
 
     class CompositeBundle(query.Bundle):
-        def __init__(self, property, expr):
-            self.property = property
+        def __init__(self, property_, expr):
+            self.property = property_
             super(CompositeProperty.CompositeBundle, self).__init__(
-                property.key, *expr)
+                property_.key, *expr
+            )
 
         def create_row_processor(self, query, procs, labels):
             def proc(row):
                 return self.property.composite_class(
-                    *[proc(row) for proc in procs])
+                    *[proc(row) for proc in procs]
+                )
+
             return proc
 
     class Comparator(PropComparator):
@@ -390,15 +401,15 @@ class CompositeProperty(DescriptorProperty):
         See the example in :ref:`composite_operations` for an overview
         of usage , as well as the documentation for :class:`.PropComparator`.
 
-        See also:
+        .. seealso::
 
-        :class:`.PropComparator`
+            :class:`.PropComparator`
 
-        :class:`.ColumnOperators`
+            :class:`.ColumnOperators`
 
-        :ref:`types_operators`
+            :ref:`types_operators`
 
-        :attr:`.TypeEngine.comparator_factory`
+            :attr:`.TypeEngine.comparator_factory`
 
         """
 
@@ -410,11 +421,13 @@ class CompositeProperty(DescriptorProperty):
 
         def __clause_element__(self):
             return expression.ClauseList(
-                group=False, *self._comparable_elements)
+                group=False, *self._comparable_elements
+            )
 
         def _query_clause_element(self):
             return CompositeProperty.CompositeBundle(
-                self.prop, self.__clause_element__())
+                self.prop, self.__clause_element__()
+            )
 
         def _bulk_update_tuples(self, value):
             if value is None:
@@ -423,22 +436,18 @@ class CompositeProperty(DescriptorProperty):
                 values = value.__composite_values__()
             else:
                 raise sa_exc.ArgumentError(
-                    "Can't UPDATE composite attribute %s to %r" %
-                    (self.prop, value))
+                    "Can't UPDATE composite attribute %s to %r"
+                    % (self.prop, value)
+                )
 
-            return zip(
-                self._comparable_elements,
-                values
-            )
+            return zip(self._comparable_elements, values)
 
         @util.memoized_property
         def _comparable_elements(self):
             if self._adapt_to_entity:
                 return [
-                    getattr(
-                        self._adapt_to_entity.entity,
-                        prop.key
-                    ) for prop in self.prop._comparable_elements
+                    getattr(self._adapt_to_entity.entity, prop.key)
+                    for prop in self.prop._comparable_elements
                 ]
             else:
                 return self.prop._comparable_elements
@@ -449,8 +458,7 @@ class CompositeProperty(DescriptorProperty):
             else:
                 values = other.__composite_values__()
             comparisons = [
-                a == b
-                for a, b in zip(self.prop._comparable_elements, values)
+                a == b for a, b in zip(self.prop._comparable_elements, values)
             ]
             if self._adapt_to_entity:
                 comparisons = [self.adapter(x) for x in comparisons]
@@ -463,7 +471,7 @@ class CompositeProperty(DescriptorProperty):
         return str(self.parent.class_.__name__) + "." + self.key
 
 
-@util.langhelpers.dependency_for("sqlalchemy.orm.properties")
+@util.langhelpers.dependency_for("sqlalchemy.orm.properties", add_to_all=True)
 class ConcreteInheritedProperty(DescriptorProperty):
     """A 'do nothing' :class:`.MapperProperty` that disables
     an attribute on a concrete subclass that is only present
@@ -493,11 +501,14 @@ class ConcreteInheritedProperty(DescriptorProperty):
 
     def __init__(self):
         super(ConcreteInheritedProperty, self).__init__()
+
         def warn():
-            raise AttributeError("Concrete %s does not implement "
-                                 "attribute %r at the instance level.  Add "
-                                 "this property explicitly to %s." %
-                                 (self.parent, self.key, self.parent))
+            raise AttributeError(
+                "Concrete %s does not implement "
+                "attribute %r at the instance level.  Add "
+                "this property explicitly to %s."
+                % (self.parent, self.key, self.parent)
+            )
 
         class NoninheritedConcreteProp(object):
             def __set__(s, obj, value):
@@ -510,15 +521,21 @@ class ConcreteInheritedProperty(DescriptorProperty):
                 if obj is None:
                     return self.descriptor
                 warn()
+
         self.descriptor = NoninheritedConcreteProp()
 
 
-@util.langhelpers.dependency_for("sqlalchemy.orm.properties")
+@util.langhelpers.dependency_for("sqlalchemy.orm.properties", add_to_all=True)
 class SynonymProperty(DescriptorProperty):
-
-    def __init__(self, name, map_column=None,
-                 descriptor=None, comparator_factory=None,
-                 doc=None, info=None):
+    def __init__(
+        self,
+        name,
+        map_column=None,
+        descriptor=None,
+        comparator_factory=None,
+        doc=None,
+        info=None,
+    ):
         """Denote an attribute name as a synonym to a mapped property,
         in that the attribute will mirror the value and expression behavior
         of another attribute.
@@ -545,11 +562,20 @@ class SynonymProperty(DescriptorProperty):
 
         :param map_column: **For classical mappings and mappings against
           an existing Table object only**.  if ``True``, the :func:`.synonym`
-          construct will locate the existing named :class:`.MapperProperty`
-          based on the attribute name of this :func:`.synonym`, and assign it
-          to a new attribute linked to the name of this :func:`.synonym`.
-          This is intended to be used with the :paramref:`.synonym.descriptor`
-          parameter::
+          construct will locate the :class:`_schema.Column`
+          object upon the mapped
+          table that would normally be associated with the attribute name of
+          this synonym, and produce a new :class:`.ColumnProperty` that instead
+          maps this :class:`_schema.Column`
+          to the alternate name given as the "name"
+          argument of the synonym; in this way, the usual step of redefining
+          the mapping of the :class:`_schema.Column`
+          to be under a different name is
+          unnecessary. This is usually intended to be used when a
+          :class:`_schema.Column`
+          is to be replaced with an attribute that also uses a
+          descriptor, that is, in conjunction with the
+          :paramref:`.synonym.descriptor` parameter::
 
             my_table = Table(
                 "my_table", metadata,
@@ -625,21 +651,23 @@ class SynonymProperty(DescriptorProperty):
 
         util.set_creation_order(self)
 
+    @property
+    def uses_objects(self):
+        return getattr(self.parent.class_, self.name).impl.uses_objects
+
     # TODO: when initialized, check _proxied_property,
     # emit a warning if its not a column-based property
 
     @util.memoized_property
     def _proxied_property(self):
         attr = getattr(self.parent.class_, self.name)
-        if not hasattr(attr, 'property') or not \
-                isinstance(attr.property, MapperProperty):
+        if not hasattr(attr, "property") or not isinstance(
+            attr.property, MapperProperty
+        ):
             raise sa_exc.InvalidRequestError(
                 """synonym() attribute "%s.%s" only supports """
-                """ORM mapped attributes, got %r""" % (
-                    self.parent.class_.__name__,
-                    self.name,
-                    attr
-                )
+                """ORM mapped attributes, got %r"""
+                % (self.parent.class_.__name__, self.name, attr)
             )
         return attr.property
 
@@ -652,48 +680,62 @@ class SynonymProperty(DescriptorProperty):
             comp = prop.comparator_factory(prop, mapper)
         return comp
 
+    def get_history(self, *arg, **kw):
+        attr = getattr(self.parent.class_, self.name)
+        return attr.impl.get_history(*arg, **kw)
+
     def set_parent(self, parent, init):
         if self.map_column:
             # implement the 'map_column' option.
-            if self.key not in parent.mapped_table.c:
+            if self.key not in parent.persist_selectable.c:
                 raise sa_exc.ArgumentError(
                     "Can't compile synonym '%s': no column on table "
                     "'%s' named '%s'"
-                    % (self.name, parent.mapped_table.description, self.key))
-            elif parent.mapped_table.c[self.key] in \
-                    parent._columntoproperty and \
-                    parent._columntoproperty[
-                parent.mapped_table.c[self.key]
-            ].key == self.name:
+                    % (
+                        self.name,
+                        parent.persist_selectable.description,
+                        self.key,
+                    )
+                )
+            elif (
+                parent.persist_selectable.c[self.key]
+                in parent._columntoproperty
+                and parent._columntoproperty[
+                    parent.persist_selectable.c[self.key]
+                ].key
+                == self.name
+            ):
                 raise sa_exc.ArgumentError(
                     "Can't call map_column=True for synonym %r=%r, "
                     "a ColumnProperty already exists keyed to the name "
-                    "%r for column %r" %
-                    (self.key, self.name, self.name, self.key)
+                    "%r for column %r"
+                    % (self.key, self.name, self.name, self.key)
                 )
-            p = properties.ColumnProperty(parent.mapped_table.c[self.key])
-            parent._configure_property(
-                self.name, p,
-                init=init,
-                setparent=True)
+            p = properties.ColumnProperty(
+                parent.persist_selectable.c[self.key]
+            )
+            parent._configure_property(self.name, p, init=init, setparent=True)
             p._mapped_by_synonym = self.key
 
         self.parent = parent
 
 
-@util.langhelpers.dependency_for("sqlalchemy.orm.properties")
+@util.langhelpers.dependency_for("sqlalchemy.orm.properties", add_to_all=True)
+@util.deprecated_cls(
+    "0.7",
+    ":func:`.comparable_property` is deprecated and will be removed in a "
+    "future release.  Please refer to the :mod:`~sqlalchemy.ext.hybrid` "
+    "extension.",
+)
 class ComparableProperty(DescriptorProperty):
     """Instruments a Python property for use in query expressions."""
 
     def __init__(
-            self, comparator_factory, descriptor=None, doc=None, info=None):
+        self, comparator_factory, descriptor=None, doc=None, info=None
+    ):
         """Provides a method of applying a :class:`.PropComparator`
         to any Python descriptor attribute.
 
-        .. versionchanged:: 0.7
-            :func:`.comparable_property` is superseded by
-            the :mod:`~sqlalchemy.ext.hybrid` extension.  See the example
-            at :ref:`hybrid_custom_comparators`.
 
         Allows any Python descriptor to behave like a SQL-enabled
         attribute when used at the class level in queries, allowing
@@ -734,7 +776,7 @@ class ComparableProperty(DescriptorProperty):
         A mapping like the above allows the ``word_insensitive`` attribute
         to render an expression like::
 
-            >>> print SearchWord.word_insensitive == "Trucks"
+            >>> print(SearchWord.word_insensitive == "Trucks")
             lower(search_word.word) = lower(:lower_1)
 
         :param comparator_factory:
